#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  Permission is hereby granted, free of charge, to any person obtaining
#  a copy of this software and associated documentation files (the
#  "Software"), to deal in the Software without restriction, including
#  without limitation the rights to use, copy, modify, merge, publish,
#  distribute, sublicense, and/or sell copies of the Software, and to
#  permit persons to whom the Software is furnished to do so, subject to
#  the following conditions:
#
#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# See the file 'CHANGES' for a list of changes
#


import array
import locale
import platform
import re
import time
import calendar

LONG_LOCALES = {
    'Unknown': 0, 'German': 1, 'English_US': 2, 'English_UK': 3, 'Spanish_Modern': 4, 'Spanish_Castilian': 5,
    'Swedish': 6, 'Finnish': 7, 'French': 8, 'French_Canadian': 9, 'Icelandic': 10, 'Italian': 11, 'Dutch': 12,
    'Norwegian': 13, 'Portuguese': 15, 'Danish': 16, 'Japanese': 17, 'Korean': 18, 'Afar': 19, 'Abkhazian': 20,
    'Afrikaans': 21, 'Amharic': 22, 'Arabic': 23, 'Assamese': 24, 'Aymara': 25, 'Azerbaijani': 26, 'Bashkir': 27,
    'Byelorussian': 28, 'Bulgarian': 29, 'Bihari': 30, 'Bislama': 31, 'Bengali': 32, 'Tibetan': 33, 'Breton': 34,
    'Catalan': 35, 'Corsican': 36, 'Czech': 37, 'Welsh': 38, 'Bhutani': 39, 'Greek': 40, 'Esperanto': 41,
    'Estonian': 42, 'Basque': 43, 'Persian': 44, 'Fiji': 45, 'Faroese': 46, 'Frisian': 47, 'Irish': 48, 'Gaelic': 49,
    'Galician': 50, 'Guarani': 51, 'Gujarati': 52, 'Hausa': 53, 'Hebrew_he': 54, 'Hindi': 55, 'Croatian': 56,
    'Hungarian': 57, 'Armenian': 58, 'Interlingua': 59, 'Indonesian': 60, 'Interlingue': 61, 'Inupiak': 62,
    'Inuktitut': 63, 'Javanese': 64, 'Georgian': 65, 'Kazakh': 66, 'Greenlandic': 67, 'Cambodian': 68, 'Kannada': 69,
    'Kashmiri': 70, 'Kurdish': 71, 'Kirghiz': 72, 'Latin': 73, 'Lingala': 74, 'Laothian': 75, 'Lithuanian': 76,
    'Latvian': 77, 'Malagasy': 78, 'Maori': 79, 'Macedonian': 80, 'Malayalam': 81, 'Mongolian': 82, 'Moldavian': 83,
    'Marathi': 84, 'Malay': 85, 'Maltese': 86, 'Burmese': 87, 'Nauru': 88, 'Nepali': 89, 'Occitan': 90, 'Oromo': 91,
    'Oriya': 92, 'Punjabi': 93, 'Polish': 94, 'Pashto': 95, 'Quechua': 96, 'Rhaeto_Romance': 97, 'Kirundi': 98,
    'Romanian': 99, 'Russian': 100, 'Kinyarwanda': 101, 'Sanskrit': 102, 'Sindhi': 103, 'Sangho': 104,
    'Serbo_Croatian': 105, 'Sinhalese': 106, 'Slovak': 107, 'Slovenian': 108, 'Samoan': 109, 'Shona': 110,
    'Somali': 111, 'Albanian': 112, 'Serbian': 113, 'Siswati': 114, 'Sesotho': 115, 'Sundanese': 116, 'Swahili': 117,
    'Tamil': 118, 'Telugu': 119, 'Tajik': 120, 'Thai': 121, 'Tigrinya': 122, 'Turkmen': 123, 'Tagalog': 124,
    'Setswana': 125, 'Tonga': 126, 'Turkish': 127, 'Tsonga': 128, 'Tatar': 129, 'Twi': 130, 'Uighur': 131,
    'Ukrainian': 132, 'Urdu': 133, 'Uzbek': 134, 'Vietnamese': 135, 'Volapuk': 136, 'Wolof': 137, 'Xhosa': 138,
    'Yiddish': 139, 'Yoruba': 140, 'Zhuang': 141, 'Chinese': 142, 'Zulu': 143,
    'Hebrew': 207, 'Norwegian_Bokmal': 214, 'Norwegian_Nynorsk': 218,
}

SHORT_LOCALES = {
    'ne': LONG_LOCALES['Nepali'], 'tr': LONG_LOCALES['Turkish'], 'da': LONG_LOCALES['Danish'],
    'gl': LONG_LOCALES['Galician'], 'my': LONG_LOCALES['Burmese'], 'ug': LONG_LOCALES['Uighur'],
    'ro': LONG_LOCALES['Romanian'], 'tn': LONG_LOCALES['Setswana'], 'ta': LONG_LOCALES['Tamil'],
    'co': LONG_LOCALES['Corsican'], 'rw': LONG_LOCALES['Kinyarwanda'], 'br': LONG_LOCALES['Breton'],
    'cy': LONG_LOCALES['Welsh'], 'bo': LONG_LOCALES['Tibetan'], 'st': LONG_LOCALES['Sesotho'],
    'ko': LONG_LOCALES['Korean'], 'mo': LONG_LOCALES['Moldavian'], 'cs': LONG_LOCALES['Czech'],
    'ps': LONG_LOCALES['Pashto'], 'km': LONG_LOCALES['Cambodian'], 'af': LONG_LOCALES['Afrikaans'],
    'is': LONG_LOCALES['Icelandic'], 'qu': LONG_LOCALES['Quechua'], 'ti': LONG_LOCALES['Tigrinya'],
    'mt': LONG_LOCALES['Maltese'], 'ky': LONG_LOCALES['Kirghiz'], 'fr_CA': LONG_LOCALES['French_Canadian'],
    'la': LONG_LOCALES['Latin'], 'hy': LONG_LOCALES['Armenian'], 'ga': LONG_LOCALES['Irish'],
    'ms': LONG_LOCALES['Malay'], 'bh': LONG_LOCALES['Bihari'], 'ka': LONG_LOCALES['Georgian'],
    'oc': LONG_LOCALES['Occitan'], 'mi': LONG_LOCALES['Maori'], 'sv': LONG_LOCALES['Swedish'],
    'it': LONG_LOCALES['Italian'], 'hu': LONG_LOCALES['Hungarian'], 'fa': LONG_LOCALES['Persian'],
    'za': LONG_LOCALES['Zhuang'], 'na': LONG_LOCALES['Nauru'], 'pt': LONG_LOCALES['Portuguese'],
    'hi': LONG_LOCALES['Hindi'], 'jw': LONG_LOCALES['Javanese'], 'ks': LONG_LOCALES['Kashmiri'],
    'ba': LONG_LOCALES['Bashkir'], 'no': LONG_LOCALES['Norwegian'], 'lv': LONG_LOCALES['Latvian'],
    'ln': LONG_LOCALES['Lingala'], 'fr': LONG_LOCALES['French'], 'id': LONG_LOCALES['Indonesian'],
    'sr': LONG_LOCALES['Serbian'], 'si': LONG_LOCALES['Sinhalese'], 'vo': LONG_LOCALES['Volapuk'],
    'om': LONG_LOCALES['Oromo'], 'ab': LONG_LOCALES['Abkhazian'], 'fi': LONG_LOCALES['Finnish'],
    'fj': LONG_LOCALES['Fiji'], 'wo': LONG_LOCALES['Wolof'], 'sn': LONG_LOCALES['Shona'], 'sd': LONG_LOCALES['Sindhi'],
    'yi': LONG_LOCALES['Yiddish'], 'ha': LONG_LOCALES['Hausa'], 'pa': LONG_LOCALES['Punjabi'],
    'sl': LONG_LOCALES['Slovenian'], 'am': LONG_LOCALES['Amharic'], 'bi': LONG_LOCALES['Bislama'],
    'mr': LONG_LOCALES['Marathi'], 'rm': LONG_LOCALES['Rhaeto_Romance'], 'dz': LONG_LOCALES['Bhutani'],
    'kn': LONG_LOCALES['Kannada'], 'rn': LONG_LOCALES['Kirundi'], 'fy': LONG_LOCALES['Frisian'],
    'eo': LONG_LOCALES['Esperanto'], 'ik': LONG_LOCALES['Inupiak'], 'mn': LONG_LOCALES['Mongolian'],
    'gd': LONG_LOCALES['Gaelic'], 'as': LONG_LOCALES['Assamese'], 'mg': LONG_LOCALES['Malagasy'],
    'tk': LONG_LOCALES['Turkmen'], 'su': LONG_LOCALES['Sundanese'], 'ru': LONG_LOCALES['Russian'],
    'ia': LONG_LOCALES['Interlingua'], 'nb': LONG_LOCALES['Norwegian_Bokmal'], 'ku': LONG_LOCALES['Kurdish'],
    'vi': LONG_LOCALES['Vietnamese'], 'az': LONG_LOCALES['Azerbaijani'], 'lo': LONG_LOCALES['Laothian'],
    'sg': LONG_LOCALES['Sangho'], 'aa': LONG_LOCALES['Afar'], 'ml': LONG_LOCALES['Malayalam'],
    'ts': LONG_LOCALES['Tsonga'], 'en_GB': LONG_LOCALES['English_UK'], 'uz': LONG_LOCALES['Uzbek'],
    'kl': LONG_LOCALES['Greenlandic'], 'iu': LONG_LOCALES['Inuktitut'], 'yo': LONG_LOCALES['Yoruba'],
    'to': LONG_LOCALES['Tonga'], 'eu': LONG_LOCALES['Basque'], 'iw': LONG_LOCALES['Hebrew'],
    'bg': LONG_LOCALES['Bulgarian'], 'gu': LONG_LOCALES['Gujarati'], 'ca': LONG_LOCALES['Catalan'],
    'pl': LONG_LOCALES['Polish'], 'sq': LONG_LOCALES['Albanian'], 'ay': LONG_LOCALES['Aymara'],
    'sk': LONG_LOCALES['Slovak'], 'uk': LONG_LOCALES['Ukrainian'], 'es': LONG_LOCALES['Spanish_Modern'],
    'sw': LONG_LOCALES['Swahili'], 'tt': LONG_LOCALES['Tatar'], 'fo': LONG_LOCALES['Faroese'],
    'or': LONG_LOCALES['Oriya'], 'ss': LONG_LOCALES['Siswati'], 'sa': LONG_LOCALES['Sanskrit'],
    'sh': LONG_LOCALES['Serbo_Croatian'], 'xh': LONG_LOCALES['Xhosa'], 'th': LONG_LOCALES['Thai'],
    'ie': LONG_LOCALES['Interlingue'], 'et': LONG_LOCALES['Estonian'], 'so': LONG_LOCALES['Somali'],
    'tl': LONG_LOCALES['Tagalog'], 'mk': LONG_LOCALES['Macedonian'], 'en': LONG_LOCALES['English_US'],
    'lt': LONG_LOCALES['Lithuanian'], 'hr': LONG_LOCALES['Croatian'], 'gn': LONG_LOCALES['Guarani'],
    'de': LONG_LOCALES['German'], 'be': LONG_LOCALES['Byelorussian'], 'zu': LONG_LOCALES['Zulu'],
    'ur': LONG_LOCALES['Urdu'], 'tw': LONG_LOCALES['Twi'], 'nn': LONG_LOCALES['Norwegian_Nynorsk'],
    'bn': LONG_LOCALES['Bengali'], 'ja': LONG_LOCALES['Japanese'], 'tg': LONG_LOCALES['Tajik'],
    'te': LONG_LOCALES['Telugu'], 'he': LONG_LOCALES['Hebrew_he'], 'zh': LONG_LOCALES['Chinese'],
    'sm': LONG_LOCALES['Samoan'], 'nl': LONG_LOCALES['Dutch'], 'ar': LONG_LOCALES['Arabic'],
    'el': LONG_LOCALES['Greek'], 'kk': LONG_LOCALES['Kazakh'], }

CHARSETS = {
    'EUC-JP': 5, 'EUC-KR': 24, 'EUC-TW': 17, 'EUROSHIFT-JIS': 44, 'IBM037': 28, 'IBM273': 29, 'IBM280': 30,
    'IBM285': 31, 'IBM297': 32, 'IBM500': 33, 'IBM930': 20, 'IBM935': 21, 'IBM937': 23, 'ISO-10646-UCS-2': 6,
    'ISO-8859-1': 2, 'ISO_8859-2': 7, 'ISO-8859-3': 8, 'ISO-8859-4': 9, 'ISO-8859-5': 10, 'ISO-8859-6': 11,
    'ISO-8859-7': 12, 'ISO-8859-8': 13, 'ISO-8859-9': 14, 'ISO-8859-10': 15, 'ISO-8859-15': 18, 'JEF': 43, 'LATIN-1': 2,
    'MACINTOSH': 3, 'MS1250': 34, 'MS1251': 35, 'MS1252': 36, 'MS1253': 37, 'MS1254': 38, 'MS1255': 39, 'MS1256': 40,
    'MS1257': 41, 'MS1258': 42, 'MS1361': 26, 'MS874': 19, 'MS932': 27, 'MS936': 22, 'MS949': 24, 'MS950': 25,
    'SHIFT-JIS': 4, 'US-ASCII': 1, 'UTF-8': 16, }

PLATFORMS = {
    'WINDOWS': 4096, 'UNIX': 8192, 'RESERVED_1': 0, 'RESERVED_2': 1, 'MS_WINDOWS': 4099, 'MACINTOSH': 16388,
    'SUNOS': 8197, 'SOLARIS': 8198, 'HP_UX': 8199, 'AIX': 8200, 'LINUX': 8201, }

DEFAULT_SEPARATOR = '((\r?\n| )+)'
TYPE_PATTERN = '^(BOOL|INT|STRING|ID|TIME|DOUBLE|UNDEFINED)$'
ATTRIBUTE_PATTERN = '^.+$'
REPEATING_PATTERN = '^(R|S)$'
BASE64_PATTERN = '^[0-9a-zA-Z+/?]+?$'
INTEGER_PATTERN = '^-?\d+$'
ENCODING_PATTERN = '^(A|H)$'
BOOLEAN_PATTERN = '^(T|F)$'
CRLF_PATTERN = '\r?\n'

SINGLE = "S"
REPEATING = "R"

BOOL = "BOOL"
INT = "INT"
STRING = "STRING"
ID = "ID"
TIME = "TIME"
DOUBLE = "DOUBLE"
UNDEFINED = "UNDEFINED"

TYPES = {
    0: BOOL,
    1: INT,
    2: STRING,
    3: ID,
    4: TIME,
    5: DOUBLE,
}

NULL_ID = "0" * 16
EMPTY_STRING = ""

MONTHS = {
    'Jan': 1, 'Feb': 2, 'Mar': 3, 'Apr': 4, 'May': 5, 'Jun': 6,
    'Jul': 7, 'Aug': 8, 'Sep': 9, 'Oct': 10, 'Nov': 11, 'Dec': 12,
}

ENCODE = {
    0: 'A', 1: 'B', 2: 'C', 3: 'D', 4: 'E', 5: 'F', 6: 'G', 7: 'H',
    8: 'I', 9: 'J', 10: 'K', 11: 'L', 12: 'M', 13: 'N', 14: 'O', 15: 'P',
    16: 'Q', 17: 'R', 18: 'S', 19: 'T', 20: 'U', 21: 'V', 22: 'W', 23: 'X',
    24: 'Y', 25: 'Z', 26: 'a', 27: 'b', 28: 'c', 29: 'd', 30: 'e', 31: 'f',
    32: 'g', 33: 'h', 34: 'i', 35: 'j', 36: 'k', 37: 'l', 38: 'm', 39: 'n',
    40: 'o', 41: 'p', 42: 'q', 43: 'r', 44: 's', 45: 't', 46: 'u', 47: 'v',
    48: 'w', 49: 'x', 50: 'y', 51: 'z', 52: '0', 53: '1', 54: '2', 55: '3',
    56: '4', 57: '5', 58: '6', 59: '7', 60: '8', 61: '9', 62: '+', 63: '/',
}

DECODE = dict((v, k) for k, v in ENCODE.items())

RPC_GET_BLOCK = 1
RPC_GET_BLOCK1 = 2
RPC_GET_BLOCK2 = 3
RPC_GET_BLOCK3 = 4
RPC_GET_BLOCK4 = 5
RPC_GET_BLOCK5 = 6
RPC_DO_PUSH = 27
RPC_NEW_SESSION_BY_ADDR = 51
RPC_CLOSE_SESSION = 52
RPC_FETCH_TYPE = 53
RPC_APPLY = 54
RPC_MULTI_NEXT = 56
RPC_CLOSE_COLLECTION = 57
RPC_APPLY_FOR_LONG = 58
RPC_APPLY_FOR_BOOL = 59
RPC_APPLY_FOR_ID = 60
RPC_APPLY_FOR_STRING = 61
RPC_APPLY_FOR_OBJECT = 62
RPC_APPLY_FOR_TIME = 63
RPC_APPLY_FOR_DOUBLE = 64
RPC_GET_NEXT_PIECE = 65

CLIENT_VERSION_STRING = "6.0.0.1 python"

DM_CLIENT_CONNECT_PROTOCOL = 2
DM_CLIENT_SESSION_RECORD_HINT = -1
DM_CLIENT_SERIALIZATION_VERSION_HINT = 2

DM_CLIENT_USE_OBDATA = 1 << 0
DM_CLIENT_USE_NEW_RPC = 1 << 1
DM_CLIENT_IS_DMCL = 1 << 2
DM_CLIENT_TZ_COMPAT = 0 << 3

CLIENT_VERSION_ARRAY = [
    0,
    DM_CLIENT_CONNECT_PROTOCOL,
    DM_CLIENT_SESSION_RECORD_HINT,
    DM_CLIENT_SERIALIZATION_VERSION_HINT,
    0,
    0,
    0,
    0,
    0,
    DM_CLIENT_USE_OBDATA
    + DM_CLIENT_USE_NEW_RPC
    + DM_CLIENT_IS_DMCL
    + DM_CLIENT_TZ_COMPAT
]

DEFAULT_BATCH_SIZE = 20

ATTRIBUTE_PREFIX = "__"

ISO8601_REGEXP = "^([0-9]){4}(-([0-9]){2}){2}T([0-9]{2}:){2}([0-9]){2}Z"


def get_platform_id():
    (system, release, version) = platform.system_alias(platform.system(), platform.release(), platform.version())
    if re.match("windows", system, re.I) or re.match("windows", release, re.I):
        return PLATFORMS['MS_WINDOWS']
    elif re.match("solaris", system, re.I) or re.match("solaris", release, re.I):
        return PLATFORMS['SOLARIS']
    elif re.match("aix", system, re.I) or re.match("aix", release, re.I):
        return PLATFORMS['AIX']
    elif re.match("hpux", system, re.I) or re.match("hpux", release, re.I):
        return PLATFORMS['HP_UX']
    elif re.match("linux", system, re.I) or re.match("linux", release, re.I):
        return PLATFORMS['LINUX']
    else:
        return 0


def get_charset_id():
    (system, release, version) = platform.system_alias(platform.system(), platform.release(), platform.version())
    data = re.split("_|\.|@?", locale.setlocale(locale.LC_ALL, ''))
    data[2] = data[2].replace("_", "-").upper()
    if data[2] in CHARSETS:
        return CHARSETS[data[2]]
    elif (re.match("windows", system, re.I)
          or re.match("windows", release, re.I)) and "MS" + data[2] in CHARSETS:
        return CHARSETS["MS" + data[2]]
    else:
        return 0


def get_locale_id():
    data = re.split("_|\.|@", locale.setlocale(locale.LC_ALL, ''))
    if data[0] + "_" + data[1] in SHORT_LOCALES:
        return SHORT_LOCALES[data[0] + "_" + data[1]]
    elif data[0] + "_" + data[1] in LONG_LOCALES:
        return LONG_LOCALES[data[0] + "_" + data[1]]
    elif data[0] in SHORT_LOCALES:
        return SHORT_LOCALES[data[0]]
    elif data[0] in LONG_LOCALES:
        return LONG_LOCALES[data[0]]
    else:
        return LONG_LOCALES['Unknown']


def get_offset_in_seconds():
    t = time.time()
    return int(time.mktime(time.gmtime(t)) - time.mktime(time.localtime(t)))


def string_to_integer_array(string):
    b = array.array("B")
    b.fromstring(string)
    return b.tolist()


def integer_array_to_string(data):
    b = array.array("B")
    b.extend(data)
    return b.tostring()


def isempty(value):
    if value is None:
        return True
    if isinstance(value, str):
        if len(value) == 0:
            return True
        elif value.isspace():
            return True
        else:
            return False
    if isinstance(value, list):
        if len(value) == 0:
            return True
        else:
            return False
    if isinstance(value, dict):
        if len(value) == 0:
            return True
        else:
            return False
    return False


def parse_address(value):
    if isempty(value):
        raise ParserException("Invalid address: %s" % value)
    if not value.startswith("INET_ADDR"):
        raise ParserException("Invalid address: %s" % value)
    chunks = value.split(" ")
    return chunks[4] + ":" + str(int(chunks[2], 16))


def parse_time(value):
    if isempty(value) or "nulldate" == value:
        return None
    if re.match(ISO8601_REGEXP, value):
        chunks = re.split("[-:TZ]", value)
        if len(chunks) != 7:
            raise ParserException("Invalid date: %s" % value)
        return calendar.timegm(
            [int(chunks[0]), int(chunks[1]), int(chunks[2]), int(chunks[3]), int(chunks[4]), int(chunks[5])])
    else:
        chunks = re.split("[: ]", value)
        if len(chunks) != 6:
            raise ParserException("Invalid date: %s" % value)
        if not chunks[0] in MONTHS:
            raise ParserException("Invalid month: %s" % chunks[0])
        return time.mktime(
            [int(chunks[5]), MONTHS[chunks[0]], int(chunks[1]), int(chunks[2]), int(chunks[3]), int(chunks[4]), 0, 0,
             -1])


def get_type_from_cache(attrName):
    return TypeCache().get(attrName)


def add_type_to_cache(typeObj):
    TypeCache().add(typeObj)


def int_to_pseudo_base64(value):
    result = ""
    while value >= 64:
        result += ENCODE.get(value % 64)
        value = (value - (value % 64)) / 64
    result += ENCODE[value]
    return result


def pseudo_base64_to_int(value):
    result = 0
    for c in list(value)[::-1]:
        if c not in DECODE:
            return None
        result = (result * 64 + DECODE[c])
    return result


class ParserException(RuntimeError):
    def __init__(self, *args, **kwargs):
        RuntimeError.__init__(self, *args, **kwargs)


class ProtocolException(RuntimeError):
    def __init__(self, *args, **kwargs):
        RuntimeError.__init__(self, *args, **kwargs)


class TypeCache:
    class __impl:

        def __init__(self):
            self.__cache = {}

        def get(self, name):
            if name in self.__cache:
                return self.__cache.get(name)
            return None

        def add(self, typeInfo):
            superType = typeInfo.super
            if superType in self.__cache and superType != "NULL":
                typeInfo.extend(self.get(superType))
            self.__cache[typeInfo.name] = typeInfo

    __instance = None

    def __init__(self):
        if not TypeCache.__instance:
            TypeCache.__instance = TypeCache.__impl()
        self.__dict__['_TypeCache__instance'] = TypeCache.__instance

    def get(self, typeName):
        return self.__instance.get(typeName)

    def add(self, typeObj):
        return self.__instance.add(typeObj)


class AttrInfo(object):
    attributes = ['position', 'name', 'type', 'repeating', 'length', 'restriction']

    def __init__(self, **kwargs):
        for attribute in AttrInfo.attributes:
            self.__setattr__(ATTRIBUTE_PREFIX + attribute, kwargs.pop(attribute, None))

    def clone(self):
        return AttrInfo(**dict((x, self.__getattr__(x)) for x in AttrInfo.attributes))

    def __getattr__(self, name):
        if name in AttrInfo.attributes:
            return self.__getattribute__(ATTRIBUTE_PREFIX + name)
        else:
            raise AttributeError("Unknown attribute %s in %s" % (name, str(self.__class__)))

    def __setattr__(self, name, value):
        if name in AttrInfo.attributes:
            AttrInfo.__setattr__(self, ATTRIBUTE_PREFIX + name, value)
        else:
            super(AttrInfo, self).__setattr__(name, value)


class AttrValue(object):
    attributes = ['name', 'type', 'length', 'repeating', 'values']

    def __init__(self, **kwargs):
        for attribute in AttrValue.attributes:
            self.__setattr__(ATTRIBUTE_PREFIX + attribute, kwargs.pop(attribute, None))
        if self.values is None:
            self.values = []
        if not isinstance(self.values, list):
            self.values = [self.values]
        if self.repeating is None:
            self.repeating = False
        if self.length is None:
            self.length = 0

    def __getattr__(self, name):
        if name in AttrValue.attributes:
            return self.__getattribute__(ATTRIBUTE_PREFIX + name)
        else:
            raise AttributeError("Unknown attribute %s in %s" % (name, str(self.__class__)))

    def __setattr__(self, name, value):
        if name in AttrValue.attributes:
            AttrValue.__setattr__(self, ATTRIBUTE_PREFIX + name, value)
        else:
            super(AttrValue, self).__setattr__(name, value)

    def __len__(self):
        if self.repeating:
            return len(self.values)
        else:
            return 1

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [self[x] for x in xrange(*key.indices(len(self)))]
        elif isinstance(key, int):
            if self.repeating:
                if key > len(self.values):
                    raise KeyError
                else:
                    return self.values[key]
            else:
                if key > 0:
                    raise KeyError
                else:
                    if len(self.values) == 0:
                        return None
                    else:
                        return self.values[0]
        else:
            raise TypeError("Invalid argument type")

    def __iter__(self):
        class iterator(object):
            def __init__(self, obj):
                self.obj = obj
                self.index = -1

            def __iter__(self):
                return self

            def next(self):
                if self.index < len(self.obj):
                    self.index += 1
                    return self.obj[self.index]
                else:
                    raise StopIteration

        return iterator(self)


class TypeInfo(object):
    attributes = ['name', 'id', 'vstamp', 'version', 'cache', 'super', 'sharedparent', 'aspectname', 'aspectshareflag',
                  'serversion']

    def __init__(self, **kwargs):
        for attribute in TypeInfo.attributes:
            self.__setattr__(ATTRIBUTE_PREFIX + attribute, kwargs.pop(attribute, None))
        self.__attrs = []
        self.__positions = {}

    def __getattr__(self, name):
        if name in TypeInfo.attributes:
            return self.__getattribute__(ATTRIBUTE_PREFIX + name)
        elif name == "attributes":
            return self.__attrs
        else:
            raise AttributeError("Unknown attribute %s in %s" % (name, str(self.__class__)))

    def __setattr__(self, name, value):
        if name in TypeInfo.attributes:
            TypeInfo.__setattr__(self, ATTRIBUTE_PREFIX + name, value)
        else:
            super(TypeInfo, self).__setattr__(name, value)

    def append(self, attrInfo):
        self.__attrs.append(attrInfo)
        if self.serversion > 0:
            if attrInfo.position > -1:
                self.__positions[attrInfo.position] = attrInfo
            elif self.name != "GeneratedType":
                raise RuntimeError("Empty position")

    def insert(self, index, attrInfo):
        self.__attrs.insert(index, attrInfo)
        if self.serversion > 0:
            if attrInfo.position > -1:
                self.__positions[attrInfo.position] = attrInfo
            elif self.name != "GeneratedType":
                raise RuntimeError("Empty position")

    def get(self, index):
        if self.serversion > 0:
            if self.name != "GeneratedType":
                return self.__positions[index]
        return self.__attrs[index]

    def count(self):
        return len(self.__attrs)

    def extend(self, type_info):
        if self.super == type_info.name:
            for i in type_info.__attrs[::-1]:
                self.insert(0, i.clone())
            self.super = type_info.super


#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#
from dctmpy import parse_address, AttrValue, STRING, INT
from dctmpy.netwise import Netwise
from dctmpy.obj.docbroker import DocbaseMap
from dctmpy.obj.typedobject import TypedObject

NETWISE_VERSION = 1
NETWISE_RELEASE = 0
NETWISE_INUMBER = 1094

version = "0.0.1 python"
handle = "localhost"


class DocbrokerClient(Netwise):
    def __init__(self, **kwargs):
        super(DocbrokerClient, self).__init__(**dict(kwargs, **{
            'version': NETWISE_VERSION,
            'release': NETWISE_RELEASE,
            'inumber': NETWISE_INUMBER,
        }))

    def get_docbasemap(self):
        return DocbaseMap(
            buffer=self._request_object(DocbrokerClient._docbasemap_request(version, handle))
        )

    def get_servermap(self, docbase):
        servermap = DocbaseMap(
            buffer=self._request_object(DocbrokerClient._servermap_request(version, handle, docbase)))
        if not 'r_host_name' in servermap:
            raise RuntimeError("No servers for docbase %s on %s" % (docbase, parse_address(servermap['i_host_addr'])))
        return servermap

    def _request_object(self, data):
        try:
            result = self.request(type=1, data=[data], immediate=True).receive().next()
        finally:
            self.disconnect()
        return result

    @staticmethod
    def _docbasemap_request(handle, version):
        obj = TypedObject(serversion=0)
        obj.add(AttrValue(name="DBR_REQUEST_NAME", type=STRING, values=["DBRN_GET_DOCBASE_MAP"]))
        obj.add(AttrValue(name="DBR_REQUEST_VERSION", type=INT, values=[1]))
        obj.add(AttrValue(name="DBR_REQUEST_HANDLE", type=STRING, values=[handle]))
        obj.add(AttrValue(name="DBR_SOFTWARE_VERSION", type=STRING, values=[version]))
        return obj

    @staticmethod
    def _servermap_request(handle, version, docbase):
        obj = TypedObject(serversion=0)
        obj.add(AttrValue(name="r_docbase_name", type=STRING, values=[docbase]))
        obj.add(AttrValue(name="r_map_name", type=STRING, values=["mn_cs_map"]))
        obj.add(AttrValue(name="DBR_REQUEST_NAME", type=STRING, values=["DBRN_GET_SERVER_MAP"]))
        obj.add(AttrValue(name="DBR_REQUEST_VERSION", type=INT, values=[1]))
        obj.add(AttrValue(name="DBR_REQUEST_HANDLE", type=STRING, values=[handle]))
        obj.add(AttrValue(name="DBR_SOFTWARE_VERSION", type=STRING, values=[version]))
        return obj





#!/usr/bin/env python

import argparse
import re
import time

try:
    from urllib.request import urlopen, URLError
except ImportError:
    from urllib2 import urlopen, URLError

from nagiosplugin import Metric, Result, Summary, Check, Resource, guarded, ScalarResult, ScalarContext
from nagiosplugin.state import Critical, Warn, Ok, Unknown

from dctmpy.docbaseclient import DocbaseClient
from dctmpy.docbrokerclient import DocbrokerClient


THRESHOLDS = 'thresholds'
NULL_CONTEXT = 'null'

JOB_ATTRIBUTES = ['object_name', 'is_inactive', 'a_last_invocation',
                  'a_last_completion', 'a_last_return_code', 'a_current_status',
                  'a_status', 'a_special_app', 'run_mode', 'run_interval',
                  'expiration_date', 'max_iterations', 'a_iterations',
                  'a_next_invocation', 'start_date', 'a_current_status']

JOB_QUERY = "SELECT " + ", ".join(JOB_ATTRIBUTES) + " FROM dm_job WHERE 1=1 "

JOB_ACTIVE_CONDITION = " AND ((a_last_invocation IS NOT NULLDATE and a_last_completion IS NULLDATE) " \
                       " OR a_special_app = 'agentexec')" \
                       " AND (i_is_reference = 0 OR i_is_reference is NULL)" \
                       " AND (i_is_replica = 0 OR i_is_replica is NULL)"

JOB_INTERVALS = {
    1: 60,
    2: 60 * 60,
    3: 24 * 60 * 60,
    4: 7 * 24 * 60 * 60
}

APP_SERVER_RESPONSES = {
    'do_method': 'Documentum Java Method Server',
    'do_mail': 'Documentum Mail Servlet',
    'do_bpm': 'Documentum Java Method Server'
}

APP_SERVER_TIMEOUT = 5


class CheckDocbase(Resource):
    def __init__(self, args, results):
        self.args = args
        self.results = results
        self.session = None

    def probe(self):
        yield CustomMetric(NULL_CONTEXT, 0, context=NULL_CONTEXT)
        try:
            self.check_login()
            if not self.session:
                return
            if self.mode == 'login':
                return
            results = modes[self.mode][0](self)
            if not results:
                return
            for result in results:
                if result:
                    yield result
        finally:
            try:
                if self.session:
                    self.session.disconnect()
            except Exception, e:
                pass

    def check_sessions(self):
        try:
            count = self.session.count_sessions()
        except Exception, e:
            self.add_result(Critical, "Unable to retrieve session count: " + str(e))
            return
        yield CustomMetric('sessioncount', int(count['hot_list_size']), min=0,
                           max=int(count['concurrent_sessions']),
                           context=THRESHOLDS)

    def check_targets(self):
        targets = []
        server_name = self.session.serverconfig['object_name']
        docbase_name = self.session.docbaseconfig['object_name']
        try:
            for target in self.session.list_targets():
                targets.extend(zip(target['projection_targets'], target['projection_ports']))
        except Exception, e:
            message = "Unable to retrieve targets: %s" % str(e)
            self.add_result(Critical, message)
            return

        for (host, port) in targets:
            self.check_registration(host, port, docbase_name, server_name)

    def check_registration(self, docbrokerhost, docbrokerport, docbaseame, servername):
        docbroker = DocbrokerClient(host=docbrokerhost, port=docbrokerport)

        try:
            docbasemap = docbroker.get_docbasemap()
        except Exception, e:
            message = "Unable to retrieve docbasemap from docbroker %s:%d: %s" % (
                docbrokerhost, docbrokerport, str(e))
            self.add_result(Critical, message)
            return

        if not docbaseame in docbasemap['r_docbase_name']:
            message = "docbase %s is not registered on %s:%d" % (docbaseame, docbrokerhost, docbrokerport)
            self.add_result(Critical, message)
            return

        try:
            servermap = docbroker.get_servermap(docbaseame)
        except Exception, e:
            message = "Unable to retrieve servermap from docbroker %s:%d: %s" % (
                docbrokerhost, docbrokerport, str(e))
            self.add_result(Critical, message)
            return

        if not servername in servermap['r_server_name']:
            message = "server %s.%s is not registered on %s:%d" % (
                docbaseame, servername, docbrokerhost, docbrokerport)
            self.add_result(Critical, message)
            return

        index = servermap['r_server_name'].index(servername)
        status = servermap['r_last_status'][index]
        docbaseid = servermap['i_docbase_id'][index]
        connaddr = servermap['i_server_connection_address'][index]

        if status != "Open":
            message = "%s.%s has status %s on %s:%d, " % (
                docbaseame, servername, status, docbrokerhost, docbrokerport)
            self.add_result(Critical, message)
            return

        chunks = connaddr.split(" ")
        host = chunks[5]
        port = int(chunks[2], 16)

        session = None
        try:
            session = DocbaseClient(host=host, port=port, docbaseid=docbaseid)
            message = "%s.%s has status %s on %s:%d" % (
                docbaseame, servername, status, docbrokerhost, docbrokerport)
            self.add_result(Ok, message)
        except Exception, e:
            message = "%s.%s has status %s on %s:%d, but error occurred during connection to %s" % (
                docbaseame, servername, status, docbrokerhost, docbrokerport, str(e))
            self.add_result(Critical, message)
            return
        if session:
            try:
                session.disconnect()
            except Exception, e:
                pass

    def check_index_agents(self):
        try:
            count = 0
            for index in CheckDocbase.get_indexes(self.session):
                count += 1
                self.check_index_agent(index['index_name'], index['object_name'])
            if count == 0:
                message = "No indexagents"
                self.add_result(Warn, message)
        except Exception, e:
            message = "Unable to execute query: %s" % str(e)
            self.add_result(Critical, message)

    def check_index_agent(self, index_name, agent_name):
        try:
            result = self.session.ftindex_agent_admin(
                index_name, agent_name
            )
            status = result['status'][0]
            if status == 0:
                message = "Indexagent %s/%s is up and running" % (index_name, agent_name)
                self.add_result(Ok, message)
            elif status == 100:
                message = "Indexagent %s/%s is stopped" % (index_name, agent_name)
                self.add_result(Warn, message)
            elif status == 200:
                message = "A problem with indexagent %s/%s" % (index_name, agent_name)
                self.add_result(Critical, message)
            else:
                message = "Indexagent %s/%s has unknown status" % (index_name, agent_name)
                self.add_result(Unknown, message)
        except Exception, e:
            message = "Unable to get indexagent %s/%s status: %s" % (
                index_name, agent_name, str(e))
            self.add_result(Critical, message)

    def check_jobs(self):
        jobs_to_check = None
        if not CheckDocbase.is_empty(self.jobs):
            if isinstance(self.jobs, list):
                jobs_to_check = list(self.jobs)
            elif isinstance(self.jobs, str):
                jobs_to_check = re.split(',\s*', self.jobs)
            else:
                raise RuntimeError("Wrong jobs argument")

        try:
            now = self.session.time()
        except Exception, e:
            message = "Unable to acquire current time: %s" % str(e)
            self.add_result(Critical, message)
            return

        try:
            for job in CheckDocbase.get_jobs(self.session, jobs_to_check):
                if jobs_to_check and job['object_name'] in jobs_to_check:
                    jobs_to_check.remove(job['object_name'])
                self.check_job(job, now)
        except Exception, e:
            message = "Unable to execute query: %s" % str(e)
            self.add_result(Critical, message)
            return

        if not CheckDocbase.is_empty(jobs_to_check):
            message = ""
            for job in jobs_to_check:
                message += "%s not found, " % job
            self.add_result(Critical, message)

    def check_job(self, job, now):
        if not job['start_date'] > -1:
            message = "%s has undefined start_date" % job['object_name']
            self.add_result(Critical, message)
            return
        if not job['a_next_invocation'] > -1:
            message = "%s has undefined next_invocation_date" % job['object_name']
            self.add_result(Critical, message)
            return
        if -1 < job['expiration_date'] < job['start_date']:
            message = "%s has expiration_date less then start_date" % job['object_name']
            self.add_result(Critical, message)
            return
        if job['max_iterations'] < 0:
            message = "%s has invalid max_iterations value: %d" % (
                (job['object_name']), (job['max_iterations']))
            self.add_result(Critical, message)
            return
        if job['run_mode'] == 0 and job['run_interval'] == 0 and job['max_iterations'] != 1:
            message = "%s has invalid max_iterations value for run_mode=0 and run_interval=0" % job[
                'object_name']
            self.add_result(Critical, message)
            return
        if job['run_mode'] in [1, 2, 3, 4] and not (1 <= job['run_interval'] <= 32767):
            message = "%s has invalid run_interval value, expected [1, 32767], got %d" % (
                (job['object_name']), (job['run_interval']))
            self.add_result(Critical, message)
            return
        if job['run_mode'] == 7 and not (-7 <= job['run_interval'] <= 7 and job['run_interval'] != 0):
            message = "%s has invalid run_interval value, expected [-7,0) U (0,7], got %d" % (
                (job['object_name']), (job['run_interval']))
            self.add_result(Critical, message)
            return
        if job['run_mode'] == 8 and not (-28 <= job['run_interval'] <= 28 and job['run_interval'] != 0):
            message = "%s has invalid run_interval value, expected [-28,0) U (0,28], got %d" % (
                (job['object_name']), (job['run_interval']))
            self.add_result(Critical, message)
            return
        if job['run_mode'] == 9 and not (-365 <= job['run_interval'] <= 365 and job['run_interval'] != 0):
            message = "%s has invalid run_interval value, expected [-365,0) U (0,365], got %d" % (
                (job['object_name']), (job['run_interval']))
            self.add_result(Critical, message)
            return
        if job['is_inactive']:
            message = "%s is inactive" % job['object_name']
            self.add_result(Critical, message)
            return
        if now > job['expiration_date'] > -1:
            message = "%s is expired" % job['object_name']
            self.add_result(Critical, message)
            return
        if 0 < job['max_iterations'] < job['a_iterations']:
            message = "%s max iterations exceeded" % job['object_name']
            self.add_result(Critical, message)
            return
        if not job['a_last_invocation'] > -1:
            message = "%s has been never executed" % job['object_name']
            self.add_result(Warn, message)
            return
        if job['a_last_return_code'] != 0:
            message = "%s has status: %s" % ((job['object_name']), (job['a_current_status']))
            self.add_result(Critical, message)
            return
        if re.match('agentexec', job['a_special_app']) or (
                    job['a_last_invocation'] > -1 and not job['a_last_completion']):
            message = "%s is running for %s" % (
                (job['object_name']), CheckDocbase.pretty_interval(now - job['a_last_invocation']))
            self.add_result(Ok, message)
            return

        time_diff = now - job['a_last_completion']

        if job['run_mode'] in [1, 2, 3, 4]:
            message = "%s last run - %s ago" % ((job['object_name']), CheckDocbase.pretty_interval(time_diff))
            if time_diff > 2 * JOB_INTERVALS[job['run_mode']] * job['run_interval']:
                self.add_result(Critical, message)
                return
            elif time_diff > JOB_INTERVALS[job['run_mode']] * job['run_interval']:
                self.add_result(Warn, message)
                return
            else:
                self.add_result(Ok, message)
                return
        else:
            message = "Scheduling type for job %s is not currently supported" % job['object_name']
            self.add_result(Critical, message)
            return

    def check_time_skew(self):
        try:
            server_time = self.session.time()
            yield CustomMetric('timeskew', abs(server_time - time.time()), min=0, context=THRESHOLDS)
        except Exception, e:
            message = "Unable to acquire current time: %s" % str(e)
            self.add_result(Critical, message)
            return

    def check_query(self):
        try:
            count = 0
            message = ""
            for rec in CheckDocbase.read_query(self.session, self.query):
                count += 1
                if count > 1:
                    message += ", "
                message += self.format.format(**rec)

            yield CustomMetric('count', count, min=0, context=THRESHOLDS).add_message(message)
        except Exception, e:
            message = "Unable to execute query: %s" % str(e)
            self.add_result(Critical, message)

    def check_count_query(self):
        try:
            result = CheckDocbase.read_object(self.session, self.query)
            yield CustomMetric('countquery', int(result.values().pop()), min=0, context=THRESHOLDS)
        except Exception, e:
            message = "Unable to execute query: %s" % str(e)
            self.add_result(Critical, message)

    def check_work_queue(self):
        query = "SELECT count(r_object_id) AS work_queue_size FROM dmi_workitem " \
                "WHERE r_runtime_state IN (0, 1) " \
                "AND r_auto_method_id > '0000000000000000' " \
                "AND a_wq_name is NULLSTRING"
        try:
            result = CheckDocbase.read_object(self.session, query)
            yield CustomMetric('workqueue', int(result['work_queue_size']), min=0, context=THRESHOLDS)
        except Exception, e:
            message = "Unable to execute query: %s" % str(e)
            self.add_result(Critical, message)

    def check_server_work_queue(self):
        server_id = self.session.serverconfig['r_object_id']
        server_name = self.session.serverconfig['object_name']
        query = "SELECT count(r_object_id) AS work_queue_size FROM dmi_workitem " \
                "WHERE r_runtime_state IN (0, 1) " \
                "AND r_auto_method_id > '0000000000000000' " \
                "AND a_wq_name ='" + server_id + "'"
        try:
            result = CheckDocbase.read_object(self.session, query)
            yield CustomMetric(server_name[-20:], int(result['work_queue_size']), min=0, context=THRESHOLDS)
        except Exception, e:
            message = "Unable to execute query: %s" % str(e)
            self.add_result(Critical, message)

    def check_fulltext_queue(self):
        try:
            count = 0
            for user in CheckDocbase.read_query(self.session,
                                                "select distinct queue_user from dm_ftindex_agent_config"):
                count += 1
                result = self.check_fulltext_queue_for_user(user['queue_user'])
                if result:
                    yield result
            if count == 0:
                message = "No indexagents"
                self.add_result(Warn, message)
        except Exception, e:
            message = "Unable to execute query: %s" % str(e)
            self.add_result(Critical, message)

    def check_fulltext_queue_for_user(self, username):
        query = "SELECT count(r_object_id) AS queue_size FROM dmi_queue_item WHERE name='" \
                + username + "'AND task_state not in ('failed','warning')"
        try:
            result = CheckDocbase.read_object(self.session, query)
            return CustomMetric(username[-20:], int(result['queue_size']), min=0, context=THRESHOLDS)
        except Exception, e:
            message = "Unable to execute query: %s" % str(e)
            self.add_result(Critical, message)

    def check_failed_tasks(self):
        try:
            count = 0
            message = ""
            for rec in CheckDocbase.get_failed_tasks(self.session):
                count += 1
                if count > 1:
                    message += ", "
                message += "'%s' (%s)" % (rec['task_name'], rec['name'])

            if count > 0:
                message = "%d task(s): %s" % (count, message)
                self.add_result(Critical, message)
        except Exception, e:
            message = "Unable to execute query: %s" % str(e)
            self.add_result(Critical, message)

    def check_jms_status(self):
        if not 'DUMP_JMS_CONFIG_LIST' in self.session.entrypoints:
            self.check_jms_status_old()
        else:
            self.check_jms_status_new()

    def check_jms_status_new(self):
        try:
            config = self.session.dump_jms_config_list()
            for i in xrange(0, len(config['jms_config_id'])):
                if config['is_disabled_in_docbase'][i]:
                    message = "%s is disabled" % config['jms_config_name'][i]
                    self.add_result(Warn, message)
                    continue
                if config['is_marked_dead_in_cache'][i]:
                    message = "%s is marked as dead" % config['jms_config_name'][i]
                    self.add_result(Critical, message)
                else:
                    try:
                        jmsconfig = self.session.fetch(config['jms_config_id'][i])
                        for i in xrange(0, len(jmsconfig['servlet_name'])):
                            name = jmsconfig['servlet_name'][i]
                            url = jmsconfig['base_uri'][i]
                            self.check_app_server(name, url)
                    except Exception, e:
                        message = "Unable to retrieve jms config %s: %s" % (config['jms_config_id'][i], str(e))
                        self.add_result(Critical, message)

        except Exception, e:
            message = "Unable to retrieve jms configs: %s" % str(e)
            self.add_result(Critical, message)
            return

    def check_jms_status_old(self):
        serverconfig = self.session.serverconfig
        for i in xrange(0, len(serverconfig['app_server_name'])):
            name = serverconfig['app_server_name'][i]
            url = serverconfig['app_server_uri'][i]
            url = re.sub(r'(https?://)(localhost(\.localdomain)?|127\.0\.0\.1)([:/])?',
                         r'\1' + serverconfig['r_host_name'] + r'\4', url)
            self.check_app_server(name, url)

    def check_app_server(self, name, url):
        if name not in APP_SERVER_RESPONSES:
            return
        try:
            response = urlopen(url, timeout=APP_SERVER_TIMEOUT)
            if response.code != 200:
                message = "unable to open %s: response code %d" % (url, response.code)
                self.add_result(Critical, message)
                return

            expected = APP_SERVER_RESPONSES[name]
            if expected not in response.read():
                message = "text \"%s\" not found in response from %s" % (expected, url)
                self.add_result(Critical, message)
                return

            message = "%s - OK" % url
            self.add_result(Ok, message)

        except URLError, e:
            message = "unable to open %s: %s" % (url, e.reason)
            self.add_result(Critical, message)
        except Exception, e:
            message = "unable to open %s: %s" % (url, str(e))
            self.add_result(Critical, message)

    def check_login(self):
        try:
            session = DocbaseClient(host=self.host, port=self.port, docbaseid=self.docbaseid)
        except Exception, e:
            message = "Unable to connect to docbase: %s" % str(e)
            self.add_result(Critical, message)
            return

        if self.login and self.authentication:
            try:
                session.authenticate(self.login, self.authentication)
            except Exception, e:
                message = "Unable to authenticate: %s" % str(e)
                self.add_result(Critical, message)
                try:
                    session.disconnect()
                except Exception, e:
                    pass
                return
            self.session = session
        else:
            message = ["No username provided", "No password provided"][not self.authentication]
            status = [Warn, Critical][self.mode != 'login']
            self.add_result(status, message)

    def add_result(self, state, message):
        self.results.add(Result(state, message))

    def __getattr__(self, name):
        if hasattr(self.args, name):
            return getattr(self.args, name)
        else:
            return AttributeError("Unknown attribute %s in %s" % (name, str(self.__class__)))

    @staticmethod
    def get_indexes(session):
        query = "select index_name, a.object_name " \
                "from dm_fulltext_index i, dm_ftindex_agent_config a " \
                "where i.index_name=a.index_name " \
                "and a.force_inactive = false"
        return CheckDocbase.read_query(session, query)

    @staticmethod
    def read_query(session, query, cnt=0):
        col = None
        try:
            read = 0
            col = session.query(query)
            for rec in col:
                yield (lambda x: dict((attr, x[attr]) for attr in x))(rec)
                read += 1
                if 0 < cnt == read:
                    break
        finally:
            if col:
                col.close()

    @staticmethod
    def read_object(session, query):
        results = CheckDocbase.read_query(session, query, 1)
        if results:
            for rec in results:
                return rec

    @staticmethod
    def get_jobs(session, jobs=None, condition=""):
        query = JOB_QUERY + condition
        if jobs:
            query += " AND object_name IN ('" + "','".join(jobs) + "')"
        return CheckDocbase.read_query(session, query)

    @staticmethod
    def get_running_jobs(session):
        return CheckDocbase.get_jobs(session, JOB_ACTIVE_CONDITION)

    @staticmethod
    def get_failed_tasks(session, offset=None):
        query = "SELECT que.task_name, que.name" \
                " FROM dmi_queue_item que, dmi_workitem wi, dmi_package pkg" \
                " WHERE que.event = 'dm_changedactivityinstancestate'" \
                " AND que.item_id LIKE '4a%%'" \
                " AND que.MESSAGE LIKE 'Activity instance, %%, of workflow, %%, failed.'" \
                " AND que.item_id = wi.r_object_id" \
                " AND wi.r_workflow_id = pkg.r_workflow_id" \
                " AND wi.r_act_seqno = pkg.r_act_seqno" \
                " AND que.delete_flag = 0"
        if offset >= 0:
            query += " que.date_sent > date(now) - %d " % offset
        return CheckDocbase.read_query(session, query)


    @staticmethod
    def pretty_interval(delta):
        if delta >= 0:
            secs = (delta) % 60
            mins = (int((delta) / 60)) % 60
            hours = (int((delta) / 3600))
            if hours < 24:
                return "%02d:%02d:%02d" % (hours, mins, secs)
            else:
                days = int(hours / 24)
                hours -= days * 24
                return "%d days %02d:%02d:%02d" % (days, hours, mins, secs)
        return "future"

    @staticmethod
    def is_empty(value):
        if value is None:
            return True
        if isinstance(value, str):
            if len(value) == 0:
                return True
            elif value.isspace():
                return True
            else:
                return False
        if isinstance(value, list):
            if len(value) == 0:
                return True
            else:
                return False
        if isinstance(value, dict):
            if len(value) == 0:
                return True
            else:
                return False
        return False


class CheckSummary(Summary):
    def verbose(self, results):
        return ''

    def ok(self, results):
        return self.format(results)

    def problem(self, results):
        return self.format(results)

    def format(self, results):
        message = ""
        for state in [Ok, Unknown, Warn, Critical]:
            hint = ", ".join(str(x) for x in results if x.state == state and not CheckDocbase.is_empty(str(x)))
            message = ", ".join(x for x in [hint, message] if not (CheckDocbase.is_empty(x)))
        return message


class CustomScalarResult(ScalarResult):
    def __str__(self):
        if hasattr(self.metric, 'message'):
            return self.metric.message
        return super(CustomScalarResult, self).__str__()


class CustomMetric(Metric):
    def add_message(self, message):
        self.message = message
        return self

    def replace(self, **attr):
        if hasattr(self, 'message'):
            return super(CustomMetric, self).replace(**attr).add_message(self.message)
        return super(CustomMetric, self).replace(**attr)


modes = {
    'sessioncount': [CheckDocbase.check_sessions, True, "checks active session count"],
    'targets': [CheckDocbase.check_targets, False, "checks whether server is registered on projection targets"],
    'indexagents': [CheckDocbase.check_index_agents, False, "checks index agent status"],
    'jobs': [CheckDocbase.check_jobs, False, "checks jobs scheduling"],
    'timeskew': [CheckDocbase.check_time_skew, True, "checks time skew between nagios host and documentum"],
    'query': [CheckDocbase.check_query, True, "checks results returned by query"],
    'countquery': [CheckDocbase.check_count_query, True, "checks results returned by query"],
    'workqueue': [CheckDocbase.check_work_queue, True, "checks workqueue size"],
    'serverworkqueue': [CheckDocbase.check_server_work_queue, True, "checks server workqueue size"],
    'indexqueue': [CheckDocbase.check_fulltext_queue, True, "checks index agent queue size"],
    'failedtasks': [CheckDocbase.check_failed_tasks, True, "checks failed tasks"],
    'login': [CheckDocbase.check_login, False, "checks login"],
    'jmsstatus': [CheckDocbase.check_jms_status, False, "checks jms status"],
}


@guarded
def main():
    argp = argparse.ArgumentParser(description=__doc__)
    argp.add_argument('-H', '--host', required=True, metavar='hostname', help='server hostname')
    argp.add_argument('-p', '--port', required=False, metavar='port', type=int, help='server port')
    argp.add_argument('-i', '--docbaseid', required=False, metavar='docbaseid', type=int, help='docbase identifier')
    argp.add_argument('-l', '--login', metavar='username', help='username')
    argp.add_argument('-a', '--authentication', metavar='password', help='password')
    argp.add_argument('-t', '--timeout', metavar='timeout', default=60, type=int,
                      help='check timeout, default is 60 seconds')
    argp.add_argument('-m', '--mode', required=True, metavar='mode',
                      help="mode to use, one of: " + ", ".join(x for x in modes.keys()))
    argp.add_argument('-j', '--jobs', metavar='jobs', default='', help='jobs to check, comma-separated list')
    argp.add_argument('-n', '--name', metavar='name', default='', help='name of check that appears in output')
    argp.add_argument('-q', '--query', metavar='query', default='', help='query to run')
    argp.add_argument('-f', '--format', metavar='format', default='', help='query output format')
    argp.add_argument('-w', '--warning', metavar='RANGE', help='warning threshold')
    argp.add_argument('-c', '--critical', metavar='RANGE', help='critical threshold')
    args = argp.parse_args()

    m = re.match('^(dctm://((.*?)(:(.*))?@)?)?([^/:]+?)(:(\d+))?(/(\d+))?$', args.host)
    if m:
        if m.group(3):
            setattr(args, 'login', m.group(3))
        if m.group(5):
            setattr(args, 'authentication', m.group(5))
        if m.group(6):
            setattr(args, 'host', m.group(6))
        if m.group(8) is not None:
            setattr(args, 'port', int(m.group(8)))
        if m.group(10) is not None:
            setattr(args, 'docbaseid', int(m.group(10)))

    if args.login and not args.authentication:
        m = re.match('^(.*?):(.*)$', args.login)
        if m:
            setattr(args, 'login', m.group(1))
            setattr(args, 'authentication', m.group(2))

    check = Check(CheckSummary())
    if args.name:
        check.name = args.name
    else:
        check.name = args.mode
    check.add(CheckDocbase(args, check.results))
    check.add(ScalarContext(THRESHOLDS, getattr(args, "warning"), getattr(args, "critical"),
                            result_cls=CustomScalarResult))
    check.main(timeout=args.timeout)


if __name__ == '__main__':
    main()


#!/usr/bin/env python

import argparse
import re

from nagiosplugin import Summary, Check, Resource, guarded, Result, Metric
from nagiosplugin.state import Critical, Warn, Ok, Unknown

from dctmpy.docbrokerclient import DocbrokerClient


NULL_CONTEXT = 'null'


class CheckDocbroker(Resource):
    def __init__(self, args, results):
        self.args = args
        self.results = results

    def probe(self):
        yield Metric(NULL_CONTEXT, 0, context=NULL_CONTEXT)
        docbroker = DocbrokerClient(host=self.host, port=self.port)
        try:
            docbasemap = docbroker.get_docbasemap()

            if not docbasemap['r_docbase_name']:
                message = "No registered docbases"
                self.add_result(Unknown, message)
                return

            if not self.docbase:
                message = "Registered docbases: " + ",".join(docbasemap['r_docbase_name'])
                self.add_result(Ok, message)
                return

            for docbase in re.split(',\s*', self.docbase):
                server = None

                if '.' in docbase:
                    chunks = docbase.split('.')
                    docbase = chunks[0]
                    server = chunks[1]

                if docbase not in docbasemap['r_docbase_name']:
                    message = "Docbase %s is not registered on %s:%d" % (docbase, self.host, self.port)
                    self.add_result(Critical, message)
                    continue

                if not server:
                    message = "Docbase %s is registered on %s:%d" % (docbase, self.host, self.port)
                    self.add_result(Ok, message)
                    continue

                try:
                    servermap = docbroker.get_servermap(docbase)
                    if server not in servermap['r_server_name']:
                        message = "Server %s.%s is not registered on %s:%d" % (docbase, server, self.host, self.port)
                        self.add_result(Critical, message)
                    else:
                        index = servermap['r_server_name'].index(server)
                        status = servermap['r_last_status'][index]
                        if status == 'Open':
                            message = "Server %s.%s is registered on %s:%d" % (
                                docbase, server, self.host, self.port)
                            self.add_result(Ok, message)
                        else:
                            message = "Server %s.%s has status %s on %s:%d" % (
                                docbase, server, status, self.host, self.port)
                            self.add_result(Ok, message)
                    continue
                except Exception, e:
                    message = "Failed to retrieve servermap for %s:" % (docbase, str(e))
                    self.add_result(Critical, message)
                    continue

        except Exception, e:
            message = "Failed to retrieve docbasemap: %s" % str(e)
            self.add_result(Critical, message)

    def add_result(self, state, message):
        self.results.add(Result(state, message))

    def __getattr__(self, name):
        if hasattr(self.args, name):
            return getattr(self.args, name)
        else:
            return AttributeError("Unknown attribute %s in %s" % (name, str(self.__class__)))

    @staticmethod
    def is_empty(value):
        if value is None:
            return True
        if isinstance(value, str):
            if len(value) == 0:
                return True
            elif value.isspace():
                return True
            else:
                return False
        if isinstance(value, list):
            if len(value) == 0:
                return True
            else:
                return False
        if isinstance(value, dict):
            if len(value) == 0:
                return True
            else:
                return False
        return False


class CheckSummary(Summary):
    def verbose(self, results):
        return ''

    def ok(self, results):
        return self.format(results)

    def problem(self, results):
        return self.format(results)

    def format(self, results):
        message = ""
        for state in [Ok, Unknown, Warn, Critical]:
            hint = ", ".join(str(x) for x in results if x.state == state and not CheckDocbroker.is_empty(str(x)))
            message = ", ".join(x for x in [hint, message] if not (CheckDocbroker.is_empty(x)))
        return message


@guarded
def main():
    argp = argparse.ArgumentParser(description=__doc__)
    argp.add_argument('-H', '--host', required=True, metavar='hostname', help='server hostname')
    argp.add_argument('-p', '--port', required=False, metavar='port', default=1489, type=int,
                      help='server port, default 1489')
    argp.add_argument('-d', '--docbase', required=False, metavar='docbase', help='docbase name')
    argp.add_argument('-n', '--name', metavar='name', default='', help='name of check that appears in output')
    argp.add_argument('-t', '--timeout', metavar='timeout', default=60, type=int,
                      help='check timeout, default is 60 seconds')
    args = argp.parse_args()

    if ':' in args.host:
        chunks = args.host.split(':')
        setattr(args, 'host', chunks[0])
        setattr(args, 'port', int(chunks[1]))

    check = Check(CheckSummary())
    if getattr(args, 'name', None):
        check.name = args.name
    check.add(CheckDocbroker(args, check.results))
    check.main(timeout=args.timeout)


if __name__ == '__main__':
    main()


#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#

from dctmpy import *
from dctmpy.obj.typedobject import TypedObject


class Persistent(TypedObject):
    def __init__(self, **kwargs):
        super(Persistent, self).__init__(**kwargs)

    def _read_type(self):
        type_name = self._next_string(ATTRIBUTE_PATTERN)
        self._next_string(ATTRIBUTE_PATTERN)
        stamp = 0
        if self.serversion > 0:
            stamp = self._read_int()
        return self.session.get_type(type_name, stamp)

    def _read_object(self):
        return super(Persistent, self)._read_object()








#  Copyright (c) 2013 Andrey B. Panfilov <andrew@panfilov.tel>
#
#  See main module for license.
#

from dctmpy.net import *
from dctmpy.net.response import Response

HEADER_SIZE = 4


class Request(object):
    attributes = ['version', 'release', 'inumber', 'sequence', 'socket', 'immediate', 'type']

    def __init__(self, **kwargs):
        for attribute in Request.attributes:
            self.__setattr__(ATTRIBUTE_PREFIX + attribute, kwargs.pop(attribute, None))

        if self.version is None or self.release is None or self.inumber is None:
            raise ProtocolException("Wrong protocol version info")

        if self.type is None:
            raise ProtocolException("Invalid request type")

        data = kwargs.pop('data', None)

        if data is None:
            self.data = None
        else:
            self.data = serialize_data(data)

        if self.immediate:
            self.send()

    def send(self):
        self.socket.sendall(
            self._build_request()
        )

    def receive(self):
        message_payload = array.array('B')
        message_payload.fromstring(self.socket.recv(HEADER_SIZE))
        if len(message_payload) == 0:
            raise ProtocolException("Unable to read header")

        message_length = 0
        for i in xrange(0, HEADER_SIZE):
            message_length = message_length << 8 | message_payload[i]

        header_payload = string_to_integer_array(self.socket.recv(2))
        if header_payload[0] != PROTOCOL_VERSION:
            raise ProtocolException("Wrong protocol 0x%X expected 0x%X" % (header_payload[0], PROTOCOL_VERSION))
        header_length = header_payload[1]

        header = string_to_integer_array(self.socket.recv(header_length))

        sequence = read_integer(header)
        if sequence != self.sequence:
            raise ProtocolException("Invalid sequence %d expected %d" % (sequence, self.sequence))

        status = read_integer(header)
        if status != 0:
            raise ProtocolException("Bad status: 0x%X" % status)

        bytes_to_read = message_length - len(header_payload) - header_length
        message = array.array('B')
        while True:
            chunk = string_to_integer_array(self.socket.recv(bytes_to_read))
            message.extend(chunk)
            if len(chunk) == 0 or len(message) == bytes_to_read:
                break

        return Response(**{
            'message': message
        })

    def _build_request(self):
        data = self._build_header()
        if self.data:
            data.extend(self.data)
        length = len(data)
        data.insert(0, length & 0x000000ff)
        data.insert(0, (length >> 8) & 0x000000ff)
        data.insert(0, (length >> 16) & 0x000000ff)
        data.insert(0, (length >> 24) & 0x000000ff)
        return data

    def _build_header(self):
        header = array.array('B')
        header.extend(serialize_integer(self.sequence))
        header.extend(serialize_integer(self.type))
        header.extend(serialize_integer(self.version))
        header.extend(serialize_integer(self.release))
        header.extend(serialize_integer(self.inumber))
        header.insert(0, len(header))
        header.insert(0, PROTOCOL_VERSION)
        return header

    def __getattr__(self, name):
        if name in Request.attributes:
            return self.__getattribute__(ATTRIBUTE_PREFIX + name)
        else:
            return AttributeError("Unknown attribute %s in %s" % (name, str(self.__class__)))

    def __setattr__(self, name, value):
        if name in Request.attributes:
            Request.__setattr__(self, ATTRIBUTE_PREFIX + name, value)
        else:
            super(Request, self).__setattr__(name, value)



